-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.outputPorts = {[1] = true}
M.deviceCategories = {shaft = true, differential = true}

local function updateVelocity(device, dt)
  device.inputAV = device[device.outputAVName] * device.gearRatio
  device.parent[device.parentOutputAVName] = device.inputAV
  for i = 1, device.numberOfMagicParents do
    device.magicParents[i][device.inputAVNames[i]] = device.inputAV
  end
end

local function updateTorque(device, dt)
  local inputTorque = device.parent[device.parentOutputTorqueName] * device.gearRatio
  for i = 1, device.numberOfMagicParents do
    inputTorque = inputTorque + device.magicParents[i][device.inputTorqueNames[i]] * device.gearRatio
  end

  device[device.outputTorqueName] = inputTorque * device.gearRatio
end

local function disconnectedUpdateVelocity(device, dt)
  --use the speed of the virtual mass, not the drivetrain on other side if the shaft is broken or disconnected, otherwise, pass through
  device.inputAV = device.virtualMassAV * device.gearRatio
  device.parent[device.parentOutputAVName] = device.inputAV
  for i = 1, device.numberOfMagicParents do
    device.magicParents[i][device.inputAVNames[i]] = device.inputAV
  end
end

local function disconnectedUpdateTorque(device, dt)
  local outputTorque = device.parent[device.parentOutputTorqueName] * device.gearRatio
  for i = 1, device.numberOfMagicParents do
    outputTorque = outputTorque + device.magicParents[i][device.inputTorqueNames[i]] * device.gearRatio
  end
  --accelerate a virtual mass with the output torque if the shaft is disconnected or broken
  device.virtualMassAV = device.virtualMassAV + outputTorque * device.invCumulativeInertia * dt
  device[device.outputTorqueName] = 0 --set to 0 to stop children receiving torque
end

local function selectUpdates(device)
  device.velocityUpdate = updateVelocity
  device.torqueUpdate = updateTorque

  if device.isBroken or device.mode == "disconnected" then
    device.velocityUpdate = disconnectedUpdateVelocity
    device.torqueUpdate = disconnectedUpdateTorque
    --make sure the virtual mass has the right AV
    device.virtualMassAV = device.inputAV
  end
end

local function validate(device)
  -- using this as initSecondStage???
  device.magicParents = {}
  device.inputTorqueNames = {}
  device.inputAVNames = {}
  for i = 1, #device.fakeParents, 1 do
    local parentName = device.fakeParents[i].name
    local parentPort = device.fakeParents[i].index
    local parent = powertrain.getDevice(parentName)
    if parent then
      if parent.outputPorts[parentPort] then
        device.numberOfMagicParents = device.numberOfMagicParents + 1
        local curIdx = device.numberOfMagicParents
        device.magicParents[curIdx] = parent
        device.inputTorqueNames[curIdx] = "outputTorque" .. tostring(parentPort)
        device.inputAVNames[curIdx] = "outputAV" .. tostring(parentPort)

        parent.children = parent.children or {} -- it had no child?
        table.insert(parent.children, device) -- injecting myself into parent
        parent.isPhysicallyDisconnected = false
        if parent.setMode then
          if (parent.type == "shaft" or parent.type == "splitShaft") and #parent.availableModes == 1 then
            parent:setMode("connected") -- fix shaft validate as PhysicallyDisconnected and sets mode to disconnected forever
            -- break their default isDisconnected state, cant care about it
          end
        end
        
        --print(string.format("linked parent %s @ outputPort %d",parentName,parentPort))
      end
    end
  end

  return true
end

local function calculateInertia(device)
  local outputInertia = 0
  local cumulativeGearRatio = nil
  local maxCumulativeGearRatio = nil
  if device.children and #device.children > 0 then
    local child = device.children[1]
    outputInertia = child.cumulativeInertia
    cumulativeGearRatio = child.cumulativeGearRatio
    maxCumulativeGearRatio = child.maxCumulativeGearRatio
  else
    --Nothing connected to this shaft :(
    outputInertia = device.virtualInertia --some default inertia
  end

  device.cumulativeInertia = outputInertia / device.gearRatio / device.gearRatio
  device.invCumulativeInertia = device.cumulativeInertia > 0 and 1 / device.cumulativeInertia or 0
  device.cumulativeGearRatio = cumulativeGearRatio * device.gearRatio
  device.maxCumulativeGearRatio = maxCumulativeGearRatio * device.gearRatio
end

local function reset(device, jbeamData)
  device.gearRatio = jbeamData.gearRatio or 1
  device.friction = jbeamData.friction or 0
  device.cumulativeInertia = 1
  device.invCumulativeInertia = 1
  device.cumulativeGearRatio = 1
  device.maxCumulativeGearRatio = 1

  device.inputAV = 0
  device.virtualMassAV = 0
  device.isBroken = false

  device[device.outputTorqueName] = 0
  device[device.outputAVName] = 0

  selectUpdates(device)
end

local function new(jbeamData)
  local device = {
    deviceCategories = shallowcopy(M.deviceCategories),
    requiredExternalInertiaOutputs = shallowcopy(M.requiredExternalInertiaOutputs),
    outputPorts = shallowcopy(M.outputPorts),
    name = jbeamData.name,
    type = jbeamData.type,
    inputName = jbeamData.inputName,
    inputIndex = jbeamData.inputIndex,
    gearRatio = jbeamData.gearRatio or 1,
    friction = jbeamData.friction or 0,
    cumulativeInertia = 1,
    invCumulativeInertia = 1,
    virtualInertia = 2,
    cumulativeGearRatio = 1,
    maxCumulativeGearRatio = 1,
    isPhysicallyDisconnected = true,
    electricsName = jbeamData.electricsName,
    visualShaftAVName = jbeamData.visualShaftAVName,
    inputAV = 0,
    virtualMassAV = 0,
    isBroken = false,
    availableModes = {"connected"},
    mode = "connected",
    defaultVirtualInertia = jbeamData.defaultVirtualInertia or nil, --meant to be nil if not specified manually
    reset = reset,
    validate = validate,
    calculateInertia = calculateInertia
  }

  local outputPortIndex = 1
  device.outputTorqueName = "outputTorque" .. tostring(outputPortIndex)
  device.outputAVName = "outputAV" .. tostring(outputPortIndex)
  device[device.outputTorqueName] = 0
  device[device.outputAVName] = 0

  device.fakeParents = jbeamData.fakeParents or {}
  device.numberOfMagicParents = 0

  selectUpdates(device)

  return device
end

M.new = new

return M
